import datetime


class BaseClass:
    def dump_object(self):
        attrs = vars(self)
        content = ""
        if len(attrs):
            content = ", ".join("%s: %s" % item for item in attrs.items())
        return content

    def raise_default_error(self, content=""):
        raise RuntimeError(content + "\n" + self.dump_object())

    def _print(self, content):
        print(
            "[%s][%s]"
            % (
                datetime.datetime.now().strftime("%d-%m-%Y-%H:%M:%S"),
                self.__class__.__name__,
            ),
            end=" ",
        )
        print(content)



from google.cloud import storage
from metadata import ModelType, platformType, input_prefix, output_prefix
from pathlib import Path
import os
from base_class import BaseClass
import json
import re
import glob
import numpy as np
import onnx


def get_traced_model(
    origin_model, example_inputs, save_path=None, model_name="default_network_name"
):
    import torch

    print("Generate jit traced model...")
    example_inputs = tuple(example_inputs.values())
    model_name = networkname_to_path(model_name)
    traced_model = torch.jit.trace(origin_model, example_inputs=example_inputs).eval()
    if save_path:
        path = save_path + "jit_traced_%s.pt" % (model_name)
        torch.jit.save(traced_model, path)
        print("%s saved." % path)
    print("Jit traced model generation success.")
    return traced_model


def get_input_info_hf(traced_model):
    shape_list = [
        (i.debugName().split(".")[0], i.type().sizes())
        for i in list(traced_model.graph.inputs())[1:]
    ]
    batch_size = shape_list[0][1][0]
    return batch_size, shape_list


def from_hf_pretrained(network_name):
    from transformers import AutoTokenizer, AutoModel

    tokenizer = AutoTokenizer.from_pretrained(
        network_name, TOKENIZERS_PARALLELISM=False
    )  # uggingface/tokenizers: The current process just got forked, after parallelism has already been used. Disabling parallelism to avoid deadlocks
    model = AutoModel.from_pretrained(network_name, return_dict=False)
    return tokenizer, model


def infer_platform_type(model_path: str):
    if model_path.startswith("gs://"):
        return platformType.GOOGLESTORAGE
    else:
        return platformType.LOCAL


def networkname_to_path(network_name):
    return network_name.replace("/", "_")


def get_bucket_object_name(url: str):
    matches = re.match("gs://(.*?)/(.*)", url)
    if matches:
        bucket, object_name = matches.groups()
    else:
        raise Exception("invalid url pattern")
    return bucket, object_name


def download_blob(bucket_name, source_blob_name, destination_file_name):
    """Downloads a blob from the bucket."""
    # The ID of your GCS bucket
    # bucket_name = "your-bucket-name"
    # The ID of your GCS object
    # source_blob_name = "storage-object-name"
    # The path to which the file should be downloaded
    # destination_file_name = "local/path/to/file"

    storage_client = storage.Client()
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(source_blob_name)
    blob.download_to_filename(destination_file_name)
    print(
        "Downloaded storage object {} from bucket {} to local file {}.".format(
            source_blob_name, bucket_name, destination_file_name
        )
    )


def upload_blob(bucket_name, source_file_name, destination_blob_name):
    """Uploads a file to the bucket."""
    # The ID of your GCS bucket
    # bucket_name = "your-bucket-name"
    # The path to your file to upload
    # source_file_name = "local/path/to/file"
    # The ID of your GCS object
    # destination_blob_name = "storage-object-name"

    storage_client = storage.Client()
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(destination_blob_name)
    blob.upload_from_filename(source_file_name)
    print(f"File {source_file_name} uploaded to {destination_blob_name}.")


def upload_blob_from_memory(bucket_name, contents, destination_blob_name):
    storage_client = storage.Client()
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(destination_blob_name)
    blob.upload_from_string(contents)
    print(
        f"{destination_blob_name} with contents {contents} has been uploaded to {bucket_name}."
    )


def upload_blobs_from_directory(
    directory_path: str, dest_bucket_name: str, dest_blob_name: str
):
    storage_client = storage.Client()
    rel_paths = glob.glob(directory_path + "/**", recursive=True)
    bucket = storage_client.bucket(dest_bucket_name)
    for local_file in rel_paths:
        remote_path = f'{dest_blob_name}/{"/".join(local_file.split(os.sep)[2:])}'
        if os.path.isfile(local_file):
            blob = bucket.blob(remote_path)
            blob.upload_from_filename(local_file)

    print(
        f"Folder: {directory_path} has been uploaded to {dest_bucket_name}/{dest_blob_name}."
    )


def download_file_from_gcp(url, dst_folder, dst_name: str):
    output_dir = Path(dst_folder)
    output_dir.mkdir(parents=True, exist_ok=True)
    bucket_name, blob_name = get_bucket_object_name(url)
    destination_file_name = f"{dst_folder}/{dst_name}"
    download_blob(bucket_name, blob_name, destination_file_name)

    return destination_file_name

def upload_outputs(bucket_name, blob_name, platform_type):
    if platform_type == platformType.GOOGLESTORAGE:
        upload_blobs_from_directory(output_prefix, bucket_name, blob_name)
        upload_blob_from_memory(bucket_name, "OK", f"{blob_name}/success")
    else:
        raise NotImplementedError


def contruct_dummy_input(input_shape, input_dtype, tensor_type, device="cuda"):
    import numpy as np

    if tensor_type == "pt":
        import torch

        dummy_input = tuple(
            [
                torch.randn(*v).type(
                    {
                        "int32": torch.int32,
                        "int64": torch.int64,
                        "float32": torch.float32,
                        "float64": torch.float64,
                    }[input_dtype[k]]
                )
                for k, v in input_shape.items()
            ]
        )
    elif tensor_type == "tf":
        import tensorflow as tf

        dummy_input = [
            tf.TensorSpec(
                v,
                {
                    "int32": tf.int32,
                    "int64": tf.int64,
                    "float32": tf.float32,
                    "float64": tf.float64,
                }[input_dtype[k]],
                name="x",
            )
            for k, v in input_shape.items()
        ]
    else:
        dummy_input = dict(
            [
                (k, np.random.rand(*v).astype(input_dtype[k]))
                for k, v in input_shape.items()
            ]
        )
    return dummy_input


def convert2onnx(platform_type, model_path, model_type, input_shape, input_dtype):
    file_path = None
    if platform_type == int(platformType.LOCAL):
        file_path = model_path
    elif platform_type == int(platformType.GOOGLESTORAGE):
        file_path = download_file_from_gcp(
            model_path, input_prefix, model_path.split("/")[-1]
        )
    else:
        raise NotImplementedError
    if file_path:
        model = None
        onnx_model = None
        if model_type == int(ModelType.PT):
            raise NotImplementedError
            # import torch
            # import sys
            # sys.path.insert(0, '/home/mac_yuan/repo/RecSys_PyTorch/saves/')
            # print(file_path)
            # model = torch.load(file_path)
            # model.eval()
            # dummy_input = contruct_dummy_input(input_shape, input_dtype, "pt")
            # torch.onnx.export(model, dummy_input, onnx_path, verbose=True)
        elif model_type == int(ModelType.TF):
            # import tensorflow as tf
            # tf.saved_model.load(file_path, tags='serve')
            # os.system("python3.9 -m tf2onnx.convert --saved-model %s --output %s" % (file_path, './inputs/model.onnx'))
            raise NotImplementedError
        elif model_type == int(ModelType.ONNX):
            pass
    if not onnx_model:
        onnx_model = onnx.load(file_path)
        os.system("cp %s inputs/model.onnx" % file_path)
    return onnx_model


class JSONConfig(BaseClass):
    def __init__(self, json_path, platform_type) -> None:
        self.load(json_path, platform_type)

    def load(self, json_path, platform_type):
        path = input_prefix + "/" + json_path
        if platform_type == platformType.GOOGLESTORAGE:
            path = download_file_from_gcp(json_path, input_prefix, "config.json")
        elif platform_type == platformType.AWSSTORAGE:
            raise NotImplementedError
        with open(path) as json_file:
            self.meta = json.load(json_file)

    def __getitem__(self, key):
        return self.meta[key]


class JSONOutput(BaseClass):
    def __init__(self, json_config: JSONConfig):
        self.meta = json_config.meta

    def save(self, file_path):
        with open(file_path, "w") as outfile:
            json.dump(self.meta, outfile)

    def __getitem__(self, key):
        return self.meta[key]

    def __setitem__(self, key, value):
        self.meta[key] = value
        self.save(output_prefix + "/output_json.json")



from ansor_engine import AnsorEngine
from base_class import BaseClass
from graph_module import GraphModuleWrapper
from utils import infer_platform_type, platformType, download_file_from_gcp
import tvm
from tvm.contrib import graph_executor

class Optimum(BaseClass):
    def __init__(self, model_name):
        self.model_name = model_name

    def run(self, onnx_model, config):
        return self._run(
            onnx_model, 
            config["target"],
            config["tuning_config"]["num_measure_trials"],
            config["tuning_config"]["mode"],
            config,
            log_file=config["tuned_log"],
            input_shape=config["model_config"]["input_shape"],
            input_dtype=config["model_config"]["input_dtype"],
            verbose=config["tuning_config"]["verbose_print"]
        )

    def _run(
        self,
        onnx_model, 
        target,
        num_measure_trials,
        mode,
        out_json,
        input_shape=None,
        input_dtype=None,
        log_file=None,
        verbose=0
    ):
        if mode == "ansor":
            ae = AnsorEngine(
                self.model_name,
                onnx_model,
                target,
                input_shape,
                input_dtype,
                out_json,
            )
            if log_file != "":
                print(
                    "Historical configuration file %s found, tuning will not be executed."
                    % log_file
                )
                ae.ansor_compile(log_file=log_file)
            else:
                ae.ansor_run_tuning(
                    num_measure_trials=num_measure_trials,
                    verbose=verbose
                )
        elif mode == "autotvm":
            raise NotImplementedError
        self.ansor_engine = ae
        self.onnx_model = onnx_model

    def get_model(self):
        return GraphModuleWrapper(self.ansor_engine.module)

    def load_model(self, input_path, target: str):
        platform_type = infer_platform_type(input_path)
        if platform_type == platformType.GOOGLESTORAGE:
            # TODO: download a directory
            download_file_from_gcp(input_path + "deploy_graph.json", "./download", "deploy_graph.json")
            download_file_from_gcp(input_path + "deploy_lib.tar", "./download", "deploy_lib.tar")
            download_file_from_gcp(input_path + "deploy_param.params", "./download", "deploy_param.params")
            input_path = "./download"
        device = tvm.device(str(target), 0)
        self._print("Load module from %s" % input_path)
        loaded_json = open(input_path + "/deploy_graph.json").read()
        loaded_lib = tvm.runtime.load_module(input_path + "/deploy_lib.tar")
        loaded_params = bytearray(
            open(input_path + "/deploy_param.params", "rb").read()
        )
        module = graph_executor.create(loaded_json, loaded_lib, device)
        module.load_params(loaded_params)
        self._print("Compile success.")
        return GraphModuleWrapper(module)



from genericpath import exists
import enum


class platformType(enum.IntEnum):
    LOCAL = 0
    GOOGLESTORAGE = 1
    AWSSTORAGE = 2


class ModelType(enum.IntEnum):
    PT = 0
    TF = 1
    ONNX = 2
    KERAS = 3


input_prefix = "./inputs"
output_prefix = "./outputs"



import numpy as np
import timeit


class GraphModuleWrapper:
    def __init__(self, module):
        self.module = module

    def __call__(self, inputs_dict):
        self.module.set_input(**inputs_dict)
        self.module.run()
        num_outputs = self.module.get_num_outputs()
        tvm_outputs = {}
        for i in range(num_outputs):
            output_name = "output_{}".format(i)
            tvm_outputs[output_name] = self.module.get_output(i).numpy()
        return tvm_outputs
    
    def predict(self, inputs_dict):
        return self.__call__(inputs_dict)



import traceback
from utils import JSONConfig, JSONOutput
import argparse

from optimum import Optimum
from utils import convert2onnx, upload_outputs, platform_type_infer
from metadata import platformType, output_prefix
from pathlib import Path

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--path",
        type=str,
        help="The path of config file in json format.",
        required=True,
    )

    args = parser.parse_args()

    config = JSONConfig(args.path, platform_type_infer(args.path))
    onnx_model = convert2onnx(
        config["platform_type"],
        config["model_path"],
        config["model_type"],
        input_shape=config["model_config"]["input_shape"],
        input_dtype=config["model_config"]["input_dtype"],
    )
    Path(output_prefix).mkdir(exist_ok=True)
    out_json = JSONOutput(config)
    out_json["status"] = 1
    try:
        optimum = Optimum(config["model_name"])
        optimum.run(
            onnx_model, 
            config
        )
    except Exception as e:
        traceback.print_exc()
        out_json["error_info"] = str(e)
        out_json["status"] = -1

    if out_json["status"] != -1:
        out_json["status"] = 4
    out_json.save(output_prefix + "/output_json.json")
    if config["need_benchmark"]:
        optimum.ansor_engine.evaluate()

    upload_outputs(
        config["output_bucket"],
        "job_id=%s" % config["job_id"],
        platformType.GOOGLESTORAGE,
    )



import io
from setuptools import find_packages
import os
from setuptools import setup

def read(fname):
    return open(fname).read()

setup(
    name="dlacc",
    version=1.4,
    url="https://gitlab.gnomondigital.com/fzyuan/dl_acceleration",
    project_urls={},
    license="Apache Software License (http://www.apache.org/licenses/LICENSE-2.0)",
    author="gnomondigital",
    author_email="contact@gnomondigital.com",
    description="A simple framework for accelerating deep learning inference runtime.",
    long_description=read("README.md"),
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=[
        "onnx",
        "onnxruntime",
        "pandas",
        "google-cloud-storage",
        "tvm"
    ],
)


from base_class import BaseClass
import os
import tvm
from tvm import auto_scheduler
import tvm.relay as relay
from tvm.contrib import graph_executor
from metadata import output_prefix, input_prefix
import numpy as np
import timeit
import onnxruntime as ort
import pandas as pd

# from tvm.contrib.debugger import debug_executor as graph_executor
from pathlib import Path

DEBUG_MODE = False


class AnsorEngine(BaseClass):
    def __init__(
        self, network_name, traced_model, target, input_shape, input_dtype, out_json
    ) -> None:
        self.network_name = network_name.replace("/", "_")
        mod, params = relay.frontend.from_onnx(
            traced_model, shape=input_shape, dtype=input_dtype
        )
        self.mod = mod
        self.params = params
        self.out_json = out_json
        self.target = target
        self.input_shape = input_shape
        self.input_dtype = input_dtype
        self.onnx_model = traced_model

    def ansor_run_tuning(
        self,
        num_measure_trials=500,
        verbose=0
    ):
        self._print("Run tuning for network=%s" % self.network_name)

        self.log_file = output_prefix + (
            "/tuninglog_network_name=%s--target=%s.json"
            % (self.network_name, str(self.target))
        )

        self._print("Extract tasks...")
        tasks, task_weights = auto_scheduler.extract_tasks(
            self.mod["main"], self.params, self.target
        )

        self._print("Begin tuning...")
        tuner = auto_scheduler.TaskScheduler(tasks, task_weights)
        tune_option = auto_scheduler.TuningOptions(
            num_measure_trials=num_measure_trials,  # change this to 20000 to achieve the best performance
            runner=auto_scheduler.LocalRunner(
                repeat=10, enable_cpu_cache_flush=True, timeout=40
            ),
            early_stopping=300,
            measure_callbacks=[auto_scheduler.RecordToFile(self.log_file)],
            verbose=verbose
        )
        use_sparse = False
        if use_sparse:
            from tvm.topi.sparse.utils import sparse_sketch_rules

            search_policy = [
                auto_scheduler.SketchPolicy(
                    task,
                    program_cost_model=auto_scheduler.XGBModel(),
                    init_search_callbacks=sparse_sketch_rules(),
                )
                for task in tasks
            ]

            tuner.tune(tune_option, search_policy=search_policy)
        else:
            tuner.tune(tune_option)
        # mark log as finished
        p = Path(self.log_file)
        name_without_extension = p.stem
        ext = p.suffix
        new_file_name = f"{name_without_extension}_finished"
        p.rename(Path(p.parent, new_file_name + ext))
        self.log_file = str(p.parent) + "/" + new_file_name + ext
        self._print("Tuning Success, configuration file saved at %s" % self.log_file)
        self.out_json["status"] = 2
        self.ansor_compile(self.log_file)
        return self

    def ansor_compile(self, log_file=None):
        output_path = output_prefix + "/optimized_model"
        if log_file:
            self.log_file = log_file
        self._print("Compile from %s" % self.log_file)
        with auto_scheduler.ApplyHistoryBest(self.log_file):
            with tvm.transform.PassContext(
                opt_level=3, config={"relay.backend.use_auto_scheduler": True}
            ):
                graph, lib, graph_params = relay.build(
                    self.mod, target=self.target, params=self.params
                )
        Path(output_path).mkdir(parents=True, exist_ok=True)
        self._save(output_path, lib, graph, graph_params)
        self.device = tvm.device(str(self.target), 0)
        if DEBUG_MODE:
            self.module = graph_executor.create(
                graph, lib, self.device, dump_root="./tvmdbg"
            )
        else:
            self.module = graph_executor.create(graph, lib, self.device)
        self._print("Compile success.")
        self.out_json["status"] = 3
        return self

    def _save(self, output_path, lib, graph, params):
        lib.export_library(output_path + "/deploy_lib.tar")
        with open(output_path + "/deploy_graph.json", "w") as fo:
            fo.write(graph)
        with open(output_path + "/deploy_param.params", "wb") as fo:
            fo.write(relay.save_param_dict(params))

    def evaluate(self):
        self._print("Evaluate inference time cost...")
        timing_results = self.module.benchmark(
            self.device, repeat=5, number=10, end_to_end=True
        )
        dummy_input = dict(
            [
                (k, np.random.rand(*v).astype(self.input_dtype[k]))
                for k, v in self.input_shape.items()
            ]
        )
        ort_sess = ort.InferenceSession(input_prefix + "/model.onnx")
        to_comp = (
            np.array(
                timeit.Timer(lambda: ort_sess.run(None, dummy_input)).repeat(
                    repeat=5, number=10
                )
            )
            / 10
        )
        prof_res = np.array(timing_results.results) * 1000
        to_comp_res = to_comp * 1000
        df_optimized = pd.DataFrame(prof_res).describe()
        df_original = pd.DataFrame(to_comp_res).describe()
        result_df = pd.concat([df_optimized, df_original], axis=1)
        result_df.columns = ["optimized", "original"]
        result_df.to_csv(output_prefix + "/inference_time.csv")
